# Define your item pipelines here
#
# Don"t forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html
import os
import re
import pymysql
import warnings
import datetime
import dataclasses
from scrapy import Request
from retrying import retry
from pymysql import cursors
from urllib.parse import urlparse
from twisted.enterprise import adbapi
from dbutils.pooled_db import PooledDB
from ayugespidertools.config import logger
from twisted.internet import defer, reactor
from scrapy.pipelines.files import FilesPipeline
from ayugespidertools.common.Params import Param
from scrapy.pipelines.images import ImagesPipeline
from ayugespidertools.MongoClient import MongoDbBase
from ayugespidertools.common.Utils import ToolsForAyu
from ayugespidertools.common.MultiPlexing import ReuseOperation


# 将 pymysql 中 Data truncated for column 警告类型置为 Error，其他警告忽略
warnings.filterwarnings("error", category=pymysql.Warning, message=".*Data truncated for column.*")


__all__ = [
    "AyuFtyMysqlPipeline",
    "AyuFtyMongoPipeline",
    "AyuTwistedMysqlPipeline",
    "AyuTurboMysqlPipeline",
    "AyuTwistedMongoPipeline",
]


class AyuMysqlPipeline:
    """
    Mysql 存储场景的 scrapy pipeline 扩展的主要功能示例
    """

    def __init__(self, table_prefix, table_enum, env, record_log_to_mysql):
        """
        初始化 Mysql 链接所需要的信息
        Args:
            table_prefix: 数据库表前缀
            table_enum: 数据表的枚举信息
            env: 当前程序部署环境名
            record_log_to_mysql: 是否需要记录程序采集的基本信息到 Mysql 中
        """
        self.table_prefix = table_prefix
        self.table_enum = table_enum
        self.env = env
        self.record_log_to_mysql = record_log_to_mysql
        # 排序规则，用于创建数据库时使用
        self.collate = None
        self.mysql_config = None
        self.conn = None
        self.cursor = None
        self.crawl_time = datetime.date.today()

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            # 数据库表前缀
            table_prefix=crawler.settings.get("MYSQL_TABLE_PREFIX", ""),
            # 数据库表枚举是否开启
            table_enum=crawler.settings.get("DATA_ENUM"),
            # 获取部署的环境
            env=crawler.settings.get("ENV"),
            # 当 record_log_to_mysql 为 True 时，会记录运行情况
            record_log_to_mysql=crawler.settings.get("RECORD_LOG_TO_MYSQL", False),
        )

    def get_column_type(self, database, table, column) -> str:
        """
        获取数据字段存储类型
        Args:
            database: 数据库
            table: 数据表
            column: 字段名称

        Returns:
            column_type: 字段存储类型
        """
        sql = '''select COLUMN_TYPE from information_schema.columns where table_schema = '%s' and 
            table_name = '%s' and COLUMN_NAME= '%s';''' % (database, table, column)
        column_type = None
        try:
            self.conn.ping(reconnect=True)
            res = self.cursor.execute(sql)
            if res:
                # 注意，此处返回的结构示例为：(('varchar(10)',),)
                lines = self.cursor.fetchall()
                column_type = lines[0][0] if len(lines) == 1 else ""
        except Exception as e:
            logger.error(f"{e}")
        return column_type

    def create_table(self, table_name: str, tabel_notes: str = "", demand_code: str = ""):
        """
        创建数据库表
        Args:
            table_name: 创建表的名称
            tabel_notes: 创建表的注释
            demand_code: 创建表的需求对应的 code 值，用于和需求中的任务对应

        Returns:
            None
        """
        # 用于表格 comment 的参数生成(即 tabel_notes 参数)
        if demand_code != "":
            tabel_notes = demand_code + "_" + tabel_notes

        sql = '''CREATE TABLE `%s` (`id` int(32) NOT NULL AUTO_INCREMENT COMMENT 'id',
            PRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=%s COLLATE=%s COMMENT='%s'; ''' % (
            table_name, self.mysql_config["charset"], self.collate, tabel_notes)
        try:
            # 执行 sql 查询，获取数据
            data = self.cursor.execute(sql)
            if any([data == 0, not data]):
                logger.info(f"创建数据表 {tabel_notes}: {table_name} 成功！")

        except Exception as e:
            logger.error(f"创建表失败，tabel_notes：{tabel_notes}，table_name：{table_name}，error：{e}")

    @retry(stop_max_attempt_number=Param.retry_num, wait_random_min=Param.retry_time_min, wait_random_max=Param.retry_time_max)
    def _connect(self, pymysql_dict_config):
        try:
            self.conn = pymysql.connect(**pymysql_dict_config)
        except Exception as e:
            logger.warning(f"目标数据库：{pymysql_dict_config['database']} 不存在，尝试创建中...")
            if "1049" in str(e):
                # 如果连接目标数据库报不存在的错误时，先创建出此目标数据库
                ReuseOperation.create_database(pymysql_dict_config)
        else:
            # 连接没有问题就直接返回连接对象
            return self.conn
        # 出现数据库不存在问题后，在创建数据库 create_database 后，再次返回连接对象
        return pymysql.connect(**pymysql_dict_config)

    def open_spider(self, spider):
        self.mysql_config = spider.mysql_config
        self.collate = ToolsForAyu.get_collate_by_charset(mysql_config=self.mysql_config)
        self.conn = self._connect(self.mysql_config)
        self.cursor = self.conn.cursor()

    def get_table_name(self, table):
        """
        组合完整的数据库表
        Args:
            table: 数据表的后缀

        Returns:
            1). 拼接成完整的数据表的值
        """
        return self.table_prefix + table

    def get_new_item(self, item):
        """
        重新整合 item
        Args:
            item: scrapy yield 的 item 信息

        Returns:
            1). 整合后的 item
        """
        new_item = {"crawl_time": self.crawl_time}
        notes_dic = {"crawl_time": "采集时间"}
        # 如果是 ayugespidertools.Items 中的各个自封装类型时
        if item.get("alldata"):
            # 如果是 Item 对象转化而来，则需要转换下，以便兼容写法
            for key, value in item["alldata"].items():
                if type(value) == dict:
                    key_value = ""
                    if value["key_value"] is not None:
                        key_value = value["key_value"]

                    new_item[key] = key_value
                    notes_dic[key] = value["notes"]
                else:
                    new_item[key] = value
                    notes_dic[key] = key

        # 兼容旧写法，直接 dict 格式的 item 即可
        else:
            # 将存入表的无关字段给去掉
            save_data_item = ReuseOperation.get_items_except_keys(dict_config=item, key_list=["table", "item_mode"])
            for k, v in save_data_item.items():
                key = k.strip()
                # 将存入表的无关字段给去掉
                if isinstance(v, str):
                    new_item[key] = v.strip()
                    notes_dic[key] = key
                else:
                    new_item[key] = v
                    notes_dic[key] = key

        return {
            "new_item": new_item,
            "notes_dic": notes_dic
        }

    def process_item(self, item, spider):
        item_dict = ToolsForAyu.convert_items_to_dict(item)
        # 先查看存储场景是否匹配
        if item_dict["item_mode"] == "Mysql":
            self.insert_item(self.get_new_item(item_dict), self.get_table_name(item_dict["table"]))
        return item

    def insert_item(self, item_o, table):
        """
        通用插入数据，将 item 数据存入 Mysql 中，item 中的 key 需要跟 Mysql 数据中的字段名称一致
        Args:
            item_o: item
            table: 数据库表名

        Returns:
            None
        """
        new_item = item_o.get("new_item")
        note_dic = item_o.get("notes_dic")
        keys = "`" + "`, `".join(new_item.keys()) + "`"
        values = ", ".join(["%s"] * len(new_item))
        sql = """INSERT INTO {table} ({keys}) values ({values}) ON DUPLICATE KEY UPDATE """.format(table=table,keys=keys, values=values)
        update = ",".join([" `{key}` = %s".format(key=key) for key in new_item])
        sql += update

        try:
            self.conn.ping(reconnect=True)
            if self.cursor.execute(sql, tuple(new_item.values()) * 2):
                self.conn.commit()

        except Exception as e:
            logger.warning(f":{e}")
            logger.warning(f"Item:{new_item}  Table: {table}")
            self.conn.rollback()
            patern_colum = re.compile(r"Unknown column '(.*?)' in 'field list'")
            if "1054" in str(e):
                text = re.findall(patern_colum, str(e))
                colum = text[0]
                notes = note_dic[colum]

                if colum == "url":
                    sql = "ALTER TABLE `%s` ADD COLUMN `%s` TEXT(500) NULL COMMENT '%s';" % (table, colum, notes)
                elif colum in ["create_time", "crawl_time"]:
                    sql = "ALTER TABLE `%s` ADD COLUMN `%s` DATE NULL DEFAULT NULL COMMENT '%s';" % (table, colum, notes)
                else:
                    sql = "ALTER TABLE `%s` ADD COLUMN `%s` VARCHAR(190) NULL DEFAULT '' COMMENT '%s';" % (table, colum, notes)

                try:
                    self.conn.ping(reconnect=True)
                    if self.cursor.execute(sql):
                        self.conn.commit()
                except Exception as e:
                    if "1060" in str(e):
                        logger.info(f"添加字段 {colum} 已存在")
                    else:
                        logger.info(f"{e}")
                return self.insert_item(item_o, table)

            elif "1146" in str(e):
                patern_table = re.compile(r"Table '(.*?)' doesn't exist")
                text = re.findall(patern_table, str(e))
                table = text[0].split(".")[1]

                # 写入表枚举
                have_create_flag = False
                if self.table_enum:
                    for name, member in self.table_enum.__members__.items():
                        table_name = self.table_prefix + member.value.get("value", "")
                        table_notes = member.value.get("notes", "")
                        demand_code = member.value.get("demand_code", "")
                        if table_name == table:
                            have_create_flag = True
                            self.create_table(table_name, table_notes, demand_code)
                            break
                        else:
                            continue

                    if have_create_flag is False:
                        self.create_table(table)

                    self.insert_item(item_o, table)
                else:
                    # 未定义 Tabel_Enum 则建表
                    logger.info("未定义数据库表枚举，进行创表操作")
                    table_name = table
                    self.create_table(table_name)
                    self.insert_item(item_o, table)

            elif "1406" in str(e):
                if "Data too long for" in str(e):
                    patern_colum = re.compile(r"Data too long for column '(.*?)' at")
                    text = re.findall(patern_colum, str(e))
                    colum = text[0]
                    notes = note_dic[colum]
                    column_type = self.get_column_type(database=self.mysql_config["database"], table=table, column=colum)
                    change_colum_type = "TEXT"
                    if "text" == column_type:
                        change_colum_type = "LONGTEXT"

                    sql = """ALTER TABLE `%s` CHANGE COLUMN `%s` `%s` %s NULL DEFAULT NULL COMMENT "%s" ;""" % (
                        table, colum, colum, change_colum_type, notes)

                    try:
                        self.conn.ping(reconnect=True)
                        if self.cursor.execute(sql):
                            self.conn.commit()
                    except Exception as e:
                        logger.info(f"更新字段类型失败 {e}")
                    return self.insert_item(item_o, table)
                else:
                    logger.info("1406 错误其他类型")

            elif "1265" in str(e):
                if "Data truncated for column" in str(e):
                    patern_colum = re.compile(r"Data truncated for column '(.*?)' at")
                    text = re.findall(patern_colum, str(e))
                    colum = text[0]
                    notes = note_dic[colum]
                    column_type = self.get_column_type(database=self.mysql_config["database"], table=table, column=colum)
                    change_colum_type = "TEXT"
                    if "text" == column_type:
                        change_colum_type = "LONGTEXT"

                    sql = """ALTER TABLE `%s` CHANGE COLUMN `%s` `%s` %s NULL DEFAULT NULL COMMENT "%s" ;""" % (
                        table, colum, colum, change_colum_type, notes)
                    try:
                        self.conn.ping(reconnect=True)
                        if self.cursor.execute(sql):
                            self.conn.commit()
                    except Exception as e:
                        logger.info(f"更新字段类型失败 {e}")
                    return self.insert_item(item_o, table)
                else:
                    logger.info("1265 错误其他类型")

            else:
                # 碰到其他的异常才打印错误日志，已处理的异常不打印
                logger.error(f"ERROR:{e}")

            try:
                self.conn.ping(reconnect=True)
            except Exception:
                self.conn = self._connect(self.mysql_config)
                self.cursor = self.conn.cursor()
                self.cursor.execute(sql, tuple(new_item.values()) * 2)
                logger.info(f"重新连接 db: =>{table}, =>{new_item} ")

    def close_spider(self, spider):
        mysql_config = spider.mysql_config
        text = {}
        stats = spider.stats.get_stats()
        error_reason = ""
        for k, v in stats.items():
            if isinstance(v, datetime.datetime):
                text[k.replace("/", "_")] = (v + datetime.timedelta(hours=8)).strftime("%Y-%m-%d %H:%M:%S")
            else:
                if "response_status_count" in k and k != "downloader/response_status_count/200":
                    status_code = k.split("/")[-1] if len(k.split("/")) > 0 else ""
                    if status_code.startswith("4"):
                        if status_code == "429":
                            error_reason += "%s错误：代理超过使用频率限制 " % status_code
                        else:
                            error_reason += "%s错误：网页失效/无此网页/网站拒绝访问 " % status_code
                    elif status_code.startswith("5"):
                        error_reason += "%s错误：网站服务器处理出错 " % status_code
                    elif status_code != "":
                        error_reason += "%s:待人工排查原因" % status_code
                elif "exception_type_count" in k:
                    error_name = k.split("/")[-1]
                    if "Timeout" in error_name:
                        error_reason += "%s:网站响应超时错误 " % error_name
                    elif "ConnectionDone" in error_name:
                        error_reason += "%s:网站与脚本连接断开 " % error_name
                    elif "ResponseNeverReceived" or "ResponseFailed" in error_name:
                        error_reason += "%s:网站无响应 " % error_name
                    else:
                        error_reason += "%s:待人工排查原因" % error_name

                text[k.replace("/", "_")] = v
        log_info = {
            "database": mysql_config["database"],
            # 脚本名称
            "spider_name": spider.name,
            # uid
            "uid": mysql_config["database"] + "|" + spider.name,
            # 请求次数统计
            "request_counts": text.get("downloader_request_count", 0),
            # 接收次数统计
            "received_count": text.get("response_received_count", 0),
            # 采集数据量
            "item_counts": text.get("item_scraped_count", 0),
            # info 数据统计
            "info_count": text.get("log_count_INFO", 0),
            # 警告数据统计
            "warning_count": text.get("log_count_WARNING", 0),
            # 错误数据统计
            "error_count": text.get("log_count_ERROR", 0),
            # 开始时间
            "start_time": text.get("start_time"),
            # 结束时间
            "finish_time": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            # 花费时间
            "spend_minutes": round((datetime.datetime.now() - stats.get("start_time") - datetime.timedelta(hours=8)).seconds / 60, 2),
            "crawl_time": self.crawl_time
        }

        # 错误原因
        if text.get("log_count_ERROR", 0):
            log_info["log_count_ERROR"] = error_reason if error_reason else "请人工排查错误原因！"

        else:
            log_info["log_count_ERROR"] = ""

        # 是否记录程序采集的基本信息到 Mysql 中
        if self.record_log_to_mysql:
            # 运行脚本统计信息
            self.insert_script_statistics(log_info)
            self.table_collection_statistics(spider_name=spider.name, database=mysql_config["database"], crawl_time=self.crawl_time)

        self.conn.close()

    def table_collection_statistics(self, spider_name: str, database: str, crawl_time: datetime.date):
        """
        统计数据库入库数据
        Args:
            spider_name: 爬虫脚本名称
            database: 数据库，保存程序采集记录保存的数据库
            crawl_time: 采集时间，程序运行时间

        Returns:
            None
        """
        sql = '''
        select concat(
            'select "',
            TABLE_name,
            '", count(id) as num , crawl_time from ',
            TABLE_SCHEMA,
            '.',
            TABLE_name,
                ' where crawl_time = "%s"'
        ) from information_schema.tables
        where TABLE_SCHEMA='%s';
        ''' % (crawl_time, database)
        table_statistics_list = []
        self.conn.ping(reconnect=True)
        self.cursor.execute(sql)
        # 获取所有记录列表
        results = self.cursor.fetchall()
        sql_list = []
        for row in results:
            sql_list.append(row[0])

        sql_all = " union all ".join(sql_list)
        self.cursor.execute(sql_all)
        results = self.cursor.fetchall()
        for row in results:
            table_statistics = dict()
            table_statistics["spider_name"] = spider_name
            table_statistics["database"] = database
            table_statistics["table_name"] = row[0]
            table_statistics["number"] = row[1]
            table_statistics["crawl_time"] = str(row[2] or crawl_time)
            table_statistics_list.append(table_statistics)
            self.insert_table_statistics(table_statistics)

    def insert_table_statistics(self, data_item):
        """
        插入统计数据到表中
        Args:
            data_item: 需要统计的入库信息

        Returns:
            None
        """
        table = "table_collection_statistics"
        self.conn.ping(reconnect=True)
        self.cursor.execute("""
            CREATE TABLE IF NOT EXISTS `{}` (
            `id` int(11) NOT NULL AUTO_INCREMENT,
            `database` varchar(255) NOT NULL DEFAULT '-' COMMENT '采集程序和记录信息存储的数据库名',
            `spider_name` varchar(255) NOT NULL DEFAULT '-' COMMENT '脚本名称',            
            `crawl_time` datetime NOT NULL COMMENT '程序运行/数据采集时间',
            `table_name` varchar(255) NOT NULL COMMENT '此项目所在库（一般某个项目放在单独的数据库中）的当前表名',
            `number` varchar(255) NOT NULL COMMENT '当前表的当前 crawl_time 的采集个数',            
            PRIMARY KEY (`id`) USING BTREE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC COMMENT='项目对应库中各表采集统计表';
        """.format(table))

        data = dict(data_item)
        keys = "`" + "`, `".join(data.keys()) + "`"
        values = ", ".join(["%s"] * len(data))
        sql = "INSERT INTO {table} ({keys}) values ({values}) ON DUPLICATE KEY UPDATE ".format(table=table, keys=keys, values=values)
        update = ",".join([" `{key}` = %s".format(key=key) for key in data])
        sql += update

        try:
            self.conn.ping(reconnect=True)
            if self.cursor.execute(sql, tuple(data.values()) * 2):
                self.conn.commit()
        except Exception as e:
            self.conn.rollback()
            logger.warning(f":{e}")

    @retry(stop_max_attempt_number=Param.retry_num, wait_random_min=Param.retry_time_min, wait_random_max=Param.retry_time_max)
    def insert_script_statistics(self, data_item):
        """
        存储运行脚本的统计信息
        Args:
            data_item: 需要插入的 log 信息

        Returns:
            None
        """
        table = "script_collection_statistics"
        self.cursor.execute("""
            CREATE TABLE IF NOT EXISTS `{}` (
            `id` int(11) NOT NULL AUTO_INCREMENT,
            `database` varchar(255) NOT NULL DEFAULT '-' COMMENT '采集程序和记录信息存储的数据库名',
            `spider_name` varchar(255) NOT NULL DEFAULT '-' COMMENT '脚本名称',
            `uid` varchar(255) NOT NULL DEFAULT '-' COMMENT 'uid',
            `request_counts` varchar(255) NOT NULL DEFAULT '-' COMMENT '请求次数统计',
            `received_count` varchar(255) NOT NULL DEFAULT '-' COMMENT '接收次数统计',
            `item_counts` varchar(255) NOT NULL DEFAULT '-' COMMENT '采集数据量',
            `info_count` varchar(255) NOT NULL DEFAULT '-' COMMENT 'info 数据统计',
            `warning_count` varchar(255) NOT NULL DEFAULT '-' COMMENT '警告数据统计',
            `error_count` varchar(255) NOT NULL DEFAULT '-' COMMENT '错误数据统计',
            `start_time` datetime NOT NULL COMMENT '开始时间',
            `finish_time` datetime NOT NULL COMMENT '结束时间',
            `spend_minutes` varchar(255) NOT NULL DEFAULT '-' COMMENT '花费时间',
            `crawl_time` datetime NOT NULL COMMENT '程序运行/数据采集时间',
            `log_count_ERROR` varchar(255) DEFAULT NULL COMMENT '错误原因',
            PRIMARY KEY (`id`) USING BTREE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC COMMENT='项目运行脚本统计信息表';
        """.format(table))

        data = dict(data_item)
        keys = "`" + "`, `".join(data.keys()) + "`"
        values = ", ".join(["%s"] * len(data))
        sql = "INSERT INTO {table} ({keys}) values ({values}) ON DUPLICATE KEY UPDATE ".format(table=table, keys=keys, values=values)
        update = ",".join([" `{key}` = %s".format(key=key) for key in data])
        sql += update
        try:
            self.conn.ping(reconnect=True)
            if self.cursor.execute(sql, tuple(data.values()) * 2):
                self.conn.commit()
        except Exception as e:
            self.conn.rollback()
            logger.warning(f":{e}")


class AyuStatsMysqlPipeline(AyuMysqlPipeline):
    """
    记录运行脚本统计信息的简单示例，不在生成中使用
    """
    def __init__(self, *args, **kwargs):
        super(AyuStatsMysqlPipeline, self).__init__(*args, **kwargs)

    def close_spider(self, spider):
        text = {}
        for k, v in spider.stats.get_stats().items():
            if isinstance(v, datetime.datetime):
                text[k.replace("/", "_")] = (v + datetime.timedelta(hours=8)).strftime("%Y-%m-%d %H:%M:%S")
            else:
                text[k.replace("/", "_")] = v

        log_info = {
            "spider_name": spider.name,
            "request_counts": text.get("downloader_request_count"),
            "received_count": text.get("response_received_count"),
            "item_counts": text.get("item_scraped_count"),
            "info_count": text.get("log_count_INFO"),
            "warning_count": text.get("log_count_WARNING"),
            "error_count": text.get("log_count_ERROR"),
            "start_time": text.get("start_time"),
            "finish_time": datetime.datetime.now(),
            "spend_minutes": (datetime.datetime.now() - spider.stats.get_stats().get("start_time") - datetime.timedelta(hours=8)).seconds / 60,
            "table": "logs"
        }
        self.insert_item(self.get_new_item(log_info), self.get_table_name(log_info.get("table")))


class AyuFtyMysqlPipeline(AyuMysqlPipeline):
    """
    Mysql 存储场景的 scrapy pipeline 扩展
    """
    def __init__(self, *args, **kwargs):
        super(AyuFtyMysqlPipeline, self).__init__(*args, **kwargs)


class AyuTurboMysqlPipeline(AyuMysqlPipeline):
    """
    Mysql 存储场景的 scrapy pipeline 扩展，使用 dbutils.pooled_db 实现
    """

    def __init__(self, pool_db_config, *args, **kwargs):
        super(AyuTurboMysqlPipeline, self).__init__(*args, **kwargs)
        self.pool_db_config = pool_db_config

    @classmethod
    def from_crawler(cls, crawler):
        pool_db_config = crawler.settings.get('POOL_DB_CONFIG', None)
        if not pool_db_config:
            logger.warning("未配置 POOL_DB_CONFIG 参数，将使用其默认参数")
            pool_db_config = {
                # 连接池允许的最大连接数
                'maxconnections': 5,
                # 连接池中空闲连接的最大数量。默认0，即无最大数量限制
                'maxcached': 0,
                # 连接的最大使用次数。默认0，即无使用次数限制
                'maxusage': 0,
                # 连接数达到最大时，新连接是否可阻塞。默认False，即达到最大连接数时，再取新连接将会报错
                'blocking': True,
            }

        return cls(
            # 数据库表前缀
            table_prefix=crawler.settings.get("MYSQL_TABLE_PREFIX", ""),
            # 数据库表枚举是否开启
            table_enum=crawler.settings.get("DATA_ENUM"),
            # 获取部署的环境
            env=crawler.settings.get("ENV"),
            # 当 record_log_to_mysql 为 True 时，会记录运行情况
            record_log_to_mysql=crawler.settings.get("RECORD_LOG_TO_MYSQL", False),

            # 数据库连接池配置
            pool_db_config=pool_db_config,
        )

    def open_spider(self, spider):
        self.mysql_config = spider.mysql_config
        self.collate = ToolsForAyu.get_collate_by_charset(mysql_config=self.mysql_config)

        # 判断目标数据库是否连接正常。若连接目标数据库错误时，创建缺失的目标数据库。这个并不需要此连接对象，直接关闭即可
        super()._connect(pymysql_dict_config=self.mysql_config).close()

        # 添加 PooledDB 的配置
        self.mysql_config.update(self.pool_db_config)
        self.conn = PooledDB(pymysql, **self.mysql_config).connection()
        # mark 为 True，查询 sql 返回字典格式数据
        # self.cursor = self.conn.cursor(pymysql.cursors.DictCursor) if self.mark else self.conn.cursor()
        self.cursor = self.conn.cursor()


class AyuTwistedMysqlPipeline(AyuMysqlPipeline):
    """
    使用 twisted 的 adbapi 实现 Mysql 存储场景下的异步操作
    注意：
        1. 推荐先用 AyuFtyMysqlPipeline 使用稳定后，再迁移至此管道
    """

    def __init__(self, *args, **kwargs):
        super(AyuTwistedMysqlPipeline, self).__init__(*args, **kwargs)
        self.dbpool = None
        # 用于建表时使用
        self.collate = None
        self.conn = None
        self.cursor = None

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            # 数据库表前缀
            table_prefix=crawler.settings.get("MYSQL_TABLE_PREFIX", ""),
            # 数据库表枚举是否开启
            table_enum=crawler.settings.get("DATA_ENUM"),
            # 获取部署的环境
            env=crawler.settings.get("ENV"),
            # 当 record_log_to_mysql 为 True 时，会记录运行情况
            record_log_to_mysql=crawler.settings.get("RECORD_LOG_TO_MYSQL", False),
        )

    def open_spider(self, spider):
        self.mysql_config = spider.mysql_config
        self.collate = ToolsForAyu.get_collate_by_charset(mysql_config=self.mysql_config)

        # 判断目标数据库是否连接正常。若连接目标数据库错误时，创建缺失的目标数据库。且需要此连接对象，不能直接关闭
        self.conn = super()._connect(pymysql_dict_config=self.mysql_config)
        self.cursor = self.conn.cursor()

        self.mysql_config['cursorclass'] = cursors.DictCursor
        self.dbpool = adbapi.ConnectionPool("pymysql", cp_reconnect=True, **self.mysql_config)
        query = self.dbpool.runInteraction(self.db_create)
        query.addErrback(self.db_create_err)

    def db_create(self, cursor):
        pass

    def db_create_err(self, failure):
        logger.error('创建数据表失败: {}'.format(failure))

    def get_column_type_by_myself(self, cursor, database, table, column) -> str:
        """
        获取数据字段存储类型
        Args:
            cursor: cursor
            database: 数据库
            table: 数据表
            column: 字段名称

        Returns:
            column_type: 字段存储类型
        """
        sql = '''select COLUMN_TYPE from information_schema.columns where table_schema = '%s' and 
            table_name = '%s' and COLUMN_NAME= '%s';''' % (database, table, column)
        column_type = None
        try:
            res = cursor.execute(sql)
            if res:
                # 注意，此处返回的结构示例为：[{'COLUMN_TYPE': 'varchar(190)'}]
                lines = cursor.fetchall()
                column_type = lines[0]["COLUMN_TYPE"] if len(lines) == 1 else ""
        except Exception as e:
            logger.error(f"{e}")
        return column_type

    def process_item(self, item, spider):
        item_dict = ToolsForAyu.convert_items_to_dict(item)
        # 先查看存储场景是否匹配
        if item_dict["item_mode"] == "Mysql":
            query = self.dbpool.runInteraction(self.db_insert, item_dict)
            query.addErrback(self.handle_error, item)
        return item

    def db_insert(self, cursor, item):
        item_o = super(AyuTwistedMysqlPipeline, self).get_new_item(item)
        table = super(AyuTwistedMysqlPipeline, self).get_table_name(item["table"])

        # 以下逻辑直接 copy 父类的 insert_item 方法，只是剔除了 commit 方法和修改 cursor 而已
        new_item = item_o.get("new_item")
        note_dic = item_o.get("notes_dic")
        keys = "`" + "`, `".join(new_item.keys()) + "`"
        values = ", ".join(["%s"] * len(new_item))
        sql = """INSERT INTO {table} ({keys}) values ({values}) ON DUPLICATE KEY UPDATE """.format(table=table, keys=keys, values=values)
        update = ",".join([" `{key}` = %s".format(key=key) for key in new_item])
        sql += update

        try:
            cursor.execute(sql, tuple(new_item.values()) * 2)

        except Exception as e:
            logger.warning(f":{e}")
            logger.warning(f"Item:{new_item}  Table: {table}")
            # self.conn.rollback()
            patern_colum = re.compile(r"Unknown column '(.*?)' in 'field list'")
            if "1054" in str(e):
                text = re.findall(patern_colum, str(e))
                colum = text[0]
                notes = note_dic[colum]

                if colum == "url":
                    sql = "ALTER TABLE `%s` ADD COLUMN `%s` TEXT(500) NULL COMMENT '%s';" % (table, colum, notes)
                elif colum in ["create_time", "crawl_time"]:
                    sql = "ALTER TABLE `%s` ADD COLUMN `%s` DATE NULL DEFAULT NULL COMMENT '%s';" % (table, colum, notes)
                else:
                    sql = "ALTER TABLE `%s` ADD COLUMN `%s` VARCHAR(190) NULL DEFAULT '' COMMENT '%s';" % (table, colum, notes)

                try:
                    cursor.execute(sql)
                except Exception as e:
                    if "1060" in str(e):
                        logger.info(f"添加字段 {colum} 已存在")
                    else:
                        logger.info(f"{e}")
                return self.db_insert(cursor, item)

            elif "1146" in str(e):
                patern_table = re.compile(r"Table '(.*?)' doesn't exist")
                text = re.findall(patern_table, str(e))
                table = text[0].split(".")[1]

                # 写入表枚举
                have_create_flag = False
                if self.table_enum:
                    for name, member in self.table_enum.__members__.items():
                        table_name = self.table_prefix + member.value.get("value", "")
                        table_notes = member.value.get("notes", "")
                        demand_code = member.value.get("demand_code", "")
                        if table_name == table:
                            have_create_flag = True
                            super(AyuTwistedMysqlPipeline, self).create_table(table_name, table_notes, demand_code)
                            break
                        else:
                            continue

                    if have_create_flag is False:
                        super(AyuTwistedMysqlPipeline, self).create_table(table)

                    self.db_insert(cursor, item)
                else:
                    # 未定义 Tabel_Enum 则建表
                    logger.info("未定义数据库表枚举，进行创表操作")
                    table_name = table
                    super(AyuTwistedMysqlPipeline, self).create_table(table_name)
                    self.db_insert(cursor, item)

            elif "1406" in str(e):
                if "Data too long for" in str(e):
                    patern_colum = re.compile(r"Data too long for column '(.*?)' at")
                    text = re.findall(patern_colum, str(e))
                    colum = text[0]
                    notes = note_dic[colum]
                    column_type = self.get_column_type_by_myself(cursor=cursor, database=self.mysql_config["database"], table=table, column=colum)
                    change_colum_type = "TEXT"
                    if "text" == column_type:
                        change_colum_type = "LONGTEXT"

                    sql = """ALTER TABLE `%s` CHANGE COLUMN `%s` `%s` %s NULL DEFAULT NULL COMMENT "%s" ;""" % (table, colum, colum, change_colum_type, notes)

                    try:
                        cursor.execute(sql)
                    except Exception as e:
                        logger.info(f"更新字段类型失败 {e}")
                    return self.db_insert(cursor, item)
                else:
                    logger.info("1406 错误其他类型")

            elif "1265" in str(e):
                if "Data truncated for column" in str(e):
                    patern_colum = re.compile(r"Data truncated for column '(.*?)' at")
                    text = re.findall(patern_colum, str(e))
                    colum = text[0]
                    notes = note_dic[colum]
                    column_type = self.get_column_type_by_myself(cursor=cursor, database=self.mysql_config["database"], table=table, column=colum)
                    change_colum_type = "TEXT"
                    if "text" == column_type:
                        change_colum_type = "LONGTEXT"

                    sql = """ALTER TABLE `%s` CHANGE COLUMN `%s` `%s` %s NULL DEFAULT NULL COMMENT "%s" ;""" % (table, colum, colum, change_colum_type, notes)
                    try:
                        cursor.execute(sql)
                    except Exception as e:
                        logger.info(f"更新字段类型失败 {e}")
                    return self.db_insert(cursor, item)
                else:
                    logger.info("1265 错误其他类型")

            else:
                # 碰到其他的异常才打印错误日志，已处理的异常不打印
                logger.error(f"ERROR:{e}")

        return item

    def handle_error(self, failure, item):
        logger.error('插入数据失败:{}, item: {}'.format(failure, item))

    def close_spider(self, spider):
        # 不删除 cursorclass 其实也不影响
        if "cursorclass" in self.mysql_config.keys():
            del self.mysql_config["cursorclass"]

        # 这里新建数据库链接，是为了正常继承父类的脚本运行统计的方法（需要 self 的 mysql 连接对象存在）
        super(AyuTwistedMysqlPipeline, self).close_spider(spider)


class FilesDownloadPipeline(FilesPipeline):
    """
    文件下载场景的 scrapy pipeline 扩展
    """

    def get_media_requests(self, item, info):
        file_url = item.get("file_url")
        if file_url:
            return Request(file_url)
        logger.info("No file_url")

    def item_completed(self, results, item, info):
        file_paths = [x["path"] for ok, x in results if ok]
        if file_paths:
            item["file_path"] = file_paths[0]
        return item

    def file_path(self, request, response=None, info=None, *, item=None):
        return "files/" + os.path.basename(urlparse(request.url).path)


class ImagesDownloadPipeline(ImagesPipeline):
    """
    图片下载场景的 scrapy pipeline 扩展
    """

    def get_media_requests(self, item, info):
        image_url = item.get("image_url")
        if image_url:
            return Request(image_url)
        logger.info("No image_url")

    def item_completed(self, results, item, info):
        image_paths = [x["path"] for ok, x in results if ok]
        if image_paths:
            item["image_path"] = image_paths[0]
        return item

    def file_path(self, request, response=None, info=None, *, item=None):
        return "images/" + os.path.basename(urlparse(request.url).path)


class AyuStatisticsMysqlPipeline:
    """
    Mysql 存储且记录脚本运行状态的简单示例
    """

    def __init__(self, env):
        self.env = env
        self.conn = None
        self.cursor = None
        self.collate = None
        self.mysql_config = None
        self.crawl_time = datetime.date.today()

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            env=crawler.settings.get("ENV", "")
        )

    @retry(stop_max_attempt_number=Param.retry_num, wait_random_min=Param.retry_time_min, wait_random_max=Param.retry_time_max)
    def _connect(self, pymysql_dict_config):
        try:
            self.conn = pymysql.connect(**pymysql_dict_config)
        except Exception as e:
            logger.warning(f"目标数据库：{pymysql_dict_config['database']} 不存在，尝试创建中...")
            if "1049" in str(e):
                ReuseOperation.create_database(pymysql_dict_config)
        else:
            # 连接没有问题就直接返回连接对象
            return self.conn
        # 出现数据库不存在问题后，在创建数据库 create_database 后，再次返回连接对象
        return pymysql.connect(**pymysql_dict_config)

    def open_spider(self, spider):
        self.mysql_config = spider.mysql_config
        self.collate = ToolsForAyu.get_collate_by_charset(mysql_config=self.mysql_config)

        self.conn = self._connect(self.mysql_config)
        self.cursor = self.conn.cursor()

    def insert(self, data_item, table):
        """
        插入数据
        Args:
            data_item: scrapy item
            table: 存储至 mysql 的表名

        Returns:
            None
        """
        # data = dict(data_item)
        data = dataclasses.asdict(data_item)
        keys = "`" + "`, `".join(data.keys()) + "`"
        values = ", ".join(["%s"] * len(data))
        sql = "INSERT INTO {table} ({keys}) values ({values}) ON DUPLICATE KEY UPDATE ".format(table=table, keys=keys, values=values)
        update = ",".join([" `{key}` = %s".format(key=key) for key in data])
        sql += update
        try:
            self.conn.ping(reconnect=True)
            if self.cursor.execute(sql, tuple(data.values()) * 2):
                self.conn.commit()
        except Exception as e:
            logger.warning(f":{e}")

    def close_spider(self, spider):
        mysql_config = spider.mysql_config

        text = {}
        stats = spider.stats.get_stats()
        error_reason = ""
        for k, v in stats.items():
            if isinstance(v, datetime.datetime):
                text[k.replace("/", "_")] = (v + datetime.timedelta(hours=8)).strftime("%Y-%m-%d %H:%M:%S")
            else:
                if "response_status_count" in k and k != "downloader/response_status_count/200":
                    status_code = k.split("/")[-1]
                    if status_code.startswith("4"):
                        if status_code == "429":
                            error_reason += "%s错误：代理超过使用频率限制 " % status_code
                        else:
                            error_reason += "%s错误：网页失效/无此网页/网站拒绝访问 " % status_code
                    elif status_code.startswith("5"):
                        error_reason += "%s错误：网站服务器处理出错 " % status_code
                    else:
                        error_reason += "%s:待人工排查原因" % status_code
                elif "exception_type_count" in k:
                    error_name = k.split("/")[-1]
                    if "Timeout" in error_name:
                        error_reason += "%s:网站响应超时错误 " % error_name
                    elif "ConnectionDone" in error_name:
                        error_reason += "%s:网站与脚本连接断开 " % error_name
                    elif "ResponseNeverReceived" or "ResponseFailed" in error_name:
                        error_reason += "%s:网站无响应 " % error_name
                    else:
                        error_reason += "%s:待人工排查原因" % error_name

                text[k.replace("/", "_")] = v
        log_info = {
            "database": mysql_config["database"],
            # 脚本名称
            "spider_name": spider.name,
            # uid
            "uid": mysql_config["database"] + "|" + spider.name,
            # 请求次数统计
            "request_counts": text.get("downloader_request_count", 0),
            # 接收次数统计
            "received_count": text.get("response_received_count", 0),
            # 采集数据量
            "item_counts": text.get("item_scraped_count", 0),
            # info 数据统计
            "info_count": text.get("log_count_INFO", 0),
            # 警告数据统计
            "warning_count": text.get("log_count_WARNING", 0),
            # 错误数据统计
            "error_count": text.get("log_count_ERROR", 0),
            # 开始时间
            "start_time": text.get("start_time"),
            # 结束时间
            "finish_time": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            # 花费时间
            "spend_minutes": round((datetime.datetime.now() - stats.get("start_time") - datetime.timedelta(hours=8)).seconds / 60, 2),
            "crawl_time": self.crawl_time
        }

        if text.get("log_count_ERROR", 0):
            # 错误原因
            log_info["log_count_ERROR"] = error_reason if error_reason else "请人工排查错误原因！"

        else:
            log_info["log_count_ERROR"] = ""

        if self.env == "prod":
            self.insert(log_info, "script_collection_statistics")

        self.conn.close()

    def process_item(self, item, spider):
        return item


class AyuFtyMongoPipeline(MongoDbBase):
    """
    MongoDB 存储场景的 scrapy pipeline 扩展
    """

    def __init__(self, mongodb_config: dict, app_conf_manage: bool):
        """
        初始化 mongoDB 连接，正常的话会返回 mongoDB 的连接对象 `connect` 和 `db` 对象
        Args:
            mongodb_config: mongDB 的连接配置
        """
        if all([not mongodb_config, not app_conf_manage]):
            raise Exception("未配置 MongoDB 连接配置！")

        self.mongodb_config = None
        # 优先从本地中取配置
        if mongodb_config:
            self.mongodb_config = ReuseOperation.dict_keys_to_lower(mongodb_config)
            super(AyuFtyMongoPipeline, self).__init__(**self.mongodb_config)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            mongodb_config=crawler.settings.get('LOCAL_MONGODB_CONFIG'),
            app_conf_manage=crawler.settings.get('APP_CONF_MANAGE'),
        )

    def open_spider(self, spider):
        if not self.mongodb_config:
            self.mongodb_config = ReuseOperation.dict_keys_to_lower(spider.mongodb_conf)
            super(AyuFtyMongoPipeline, self).__init__(**self.mongodb_config)

        # 用于输出日志
        if all([self.conn, self.db]):
            logger.info(f"已连接至 host: {self.mongodb_config['host']}, database: {self.mongodb_config['database']} 的 MongoDB 目标数据库")

    def close_spider(self, spider):
        self.conn.close()

    def process_item(self, item, spider):
        """
        mongoDB 存储的方法，item["mongo_update_rule"] 用于存储查询条件，如果查询数据存在的话就更新，不存在的话就插入；
        如果没有 mongo_update_rule 则每次都新增
        Args:
            item: scrapy item
            spider: scrapy spider

        Returns:
            item: scrapy item
        """
        item_dict = ToolsForAyu.convert_items_to_dict(item)
        # 先查看存储场景是否匹配
        if item_dict["item_mode"] == "MongoDB":
            insert_data = item_dict.get("alldata")
            # 如果有 alldata 字段，则其为推荐格式
            if insert_data:
                # 判断数据中中的 alldata 的格式：(1.推荐：是嵌套 dict，就像 AyuMysqlPipeline 一样 -- 这是为了通用写法风格；2. 是单层的 dict)
                # 如果是嵌套格式的话，需要再转化为正常格式，因为此场景不需要像 Mysql 一样依赖备注来生成字段注释
                if any([isinstance(v, dict) for v in insert_data.values()]):
                    insert_data = {v: insert_data[v]["key_value"] for v in insert_data.keys()}

            # 否则为旧格式
            else:
                insert_data = ReuseOperation.get_items_except_keys(dict_config=item_dict, key_list=["table", "item_mode", "mongo_update_rule"])

            # 如果没有查重字段时，就直接插入数据（不去重）
            if not item_dict.get("mongo_update_rule"):
                self.db[item_dict["table"]].insert(insert_data)
            else:
                self.db[item_dict["table"]].update(item_dict["mongo_update_rule"], {'$set': dict(insert_data)}, True)
        return item


class AyuTwistedMongoPipeline(AyuFtyMongoPipeline):
    """
    使用 twisted 的 adbapi 实现 mongoDB 存储场景下的异步操作
    """

    def __init__(self, *args, **kwargs):
        super(AyuTwistedMongoPipeline, self).__init__(*args, **kwargs)

    def spider_closed(self, spider):
        self.conn.close()

    @defer.inlineCallbacks
    def process_item(self, item, spider):
        out = defer.Deferred()
        reactor.callInThread(self.db_insert, item, out)
        yield out
        defer.returnValue(item)

    def db_insert(self, item, out):
        item_dict = ToolsForAyu.convert_items_to_dict(item)
        # 先查看存储场景是否匹配
        if item_dict["item_mode"] == "MongoDB":
            insert_data = item_dict.get("alldata")
            # 如果有 alldata 字段，则其为推荐格式
            if insert_data:
                if any([isinstance(v, dict) for v in insert_data.values()]):
                    insert_data = {v: insert_data[v]["key_value"] for v in insert_data.keys()}

            # 否则为旧格式
            else:
                insert_data = ReuseOperation.get_items_except_keys(dict_config=item_dict, key_list=["table", "item_mode", "mongo_update_rule"])

            self.db[item_dict["table"]].update(item_dict["mongo_update_rule"], {'$set': dict(insert_data)}, True)
            reactor.callFromThread(out.callback, item_dict)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  DownloaderMiddlewares.py
@Time    :  2022/8/30 16:36
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import asyncio
import aiohttp
import urllib.parse
from loguru import logger
from scrapy.http import HtmlResponse
from scrapy.utils.defer import deferred_to_future
from twisted.internet.defer import Deferred
from ayugespidertools.common.Params import Param
from ayugespidertools.common.Utils import ToolsForAyu
from ayugespidertools.common.MultiPlexing import ReuseOperation


__all__ = [
    "AiohttpMiddleware",
    "AiohttpAsyncMiddleware",
]


def as_deferred(f):
    """
    transform a Twisted Deffered to an Asyncio Future
    Args:
        f: async function

    Returns:
        1).Deferred
    """
    return Deferred.fromFuture(asyncio.ensure_future(f))


class AiohttpMiddleware(object):
    """
    Downloader middleware handling the requests with aiohttp
    """

    def _retry(self, request, reason, spider):
        """
        TODO: 重试方法，后续添加
        """
        pass

    @classmethod
    def from_crawler(cls, crawler):
        """
        初始化 middleware
        """
        settings = crawler.settings
        # 自定义 aiohttp 全局配置信息，优先级小于 aiohttp_meta 中的配置
        local_aiohttp_conf = settings.get("LOCAL_AIOHTTP_CONFIG", {})

        # 这里的配置信息如果在 aiohttp_meta 中重复设置，则会更新当前请求的参数
        if ReuseOperation.if_dict_meet_min_limit(dict_config=local_aiohttp_conf, key_list=["TIMEOUT", "RETRY_TIMES"]):
            aiohttp_conf_temp = {
                "TIMEOUT": local_aiohttp_conf.get("TIMEOUT"),
                "SLEEP": local_aiohttp_conf.get("SLEEP"),
                "PROXY": local_aiohttp_conf.get("PROXY"),
                "RETRY_TIMES": local_aiohttp_conf.get("RETRY_TIMES"),
            }

            aiohttp_conf_lowered = ReuseOperation.dict_keys_to_lower(aiohttp_conf_temp)

            # 初始化所需要的参数信息，用于构建 aiohttp 请求信息
            # cls.download_timeout = settings.get('DOWNLOAD_TIMEOUT', settings.get('DOWNLOAD_TIMEOUT', 5))
            # cls.sleep = settings.get('SLEEP', 1)
            # cls.retry_enabled = settings.getbool('RETRY_ENABLED')
            # cls.max_retry_times = settings.getint('RETRY_TIMES')
            # cls.retry_http_codes = set(int(x) for x in settings.getlist('RETRY_HTTP_CODES'))
            # cls.proxy = settings.get('PROXY')
            cls.timeout = aiohttp_conf_lowered.get("timeout", None)
            cls.proxy = aiohttp_conf_lowered.get("proxy", None)
            cls.sleep = aiohttp_conf_lowered.get("sleep", None)
            cls.retry_times = aiohttp_conf_lowered.get("retry_times", None)
            # 用来标记是否在配置中设置
            cls.aio_local_conf_exists = True

        else:
            logger.warning(f"aiohttp 运行缺少最低配置内容要求，示例为：{Param.aiohttp_conf_example}，现尝试从 aiohttp_args 中获取！")

        cls.aio_local_conf_exists = False
        return cls()

    async def _request_by_aiohttp(self, aio_request_args: dict, aio_request_cookies: dict) -> str:
        async with aiohttp.ClientSession(cookies=aio_request_cookies) as session:
            async with session.request(**aio_request_args) as response:
                # logger.info(f"Status: {response.status}")
                # logger.info(f"Content-type: {response.headers['content-type']}")
                html = await response.text()
                return html

    async def _process_request(self, request, spider):
        """
        使用 aiohttp 来 process spider
        """
        aiohttp_meta = request.meta.get('aiohttp_args', {})
        if not isinstance(aiohttp_meta, dict):
            raise Exception("aiohttp_args 参数的格式不是 dict，请查看！")

        # set proxy

        # set cookies domain 参数
        parse_result = urllib.parse.urlsplit(request.url)
        domain = parse_result.hostname

        # _timeout = self.download_timeout
        # if aiohttp_meta.get('timeout') is not None:
        #     _timeout = aiohttp_meta.get('timeout')

        # logger.debug(f'crawling {request.url}')
        html_content = None
        try:
            # 获取请求头信息
            aiohttp_headers = ToolsForAyu.get_dict_form_scrapy_req_headers(scrapy_headers=request.headers)
            aiohttp_headers_args = {"headers": aiohttp_headers}

            # 确定请求方式
            scrapy_request_method = str(request.method).upper()
            # 默认请求方式为 get
            default_aiohttp_args_dict = {
                "method": "GET",
                "url": request.url,
            }
            if scrapy_request_method == "GET":
                aiohttp_args_dict = default_aiohttp_args_dict

            elif scrapy_request_method == "POST":
                aiohttp_args_dict = {
                    "method": "POST",
                    "url": request.url,
                }

            else:
                logger.warning(f"出现未知请求方式，请及时查看，默认 GET")
                aiohttp_args_dict = default_aiohttp_args_dict

            # 设置请求 body 参数，GET 情况下和 POST 情况下的请求参数处理
            request_body_str = str(request.body, encoding="utf-8")

            if request_body_str:
                # 如果是 json 字典格式的数据时，则是 scrapy body 传来的 json dumps 参数
                if ReuseOperation.judge_str_is_json(judge_str=request_body_str):
                    aiohttp_args_dict["data"] = request_body_str

                # 否则就是传来的字典格式
                else:
                    req_body_dict = ReuseOperation.get_req_dict_from_scrapy(req_body_data_str=request_body_str)
                    aiohttp_args_dict["data"] = req_body_dict

            # 如果存在请求头，则设置此请求头
            if aiohttp_headers:
                aiohttp_args_dict.update(aiohttp_headers_args)

            # 设置 cookies，优先从 AiohttpRequest 中的 cookies 参数中取值，没有时再从 headers 中取值
            aiohttp_cookie_dict = request.cookies
            # logger.debug(f"使用 cookies 参数中 ck 的值: {aiohttp_cookie_dict}")
            if all([not aiohttp_cookie_dict, request.headers.get("Cookie", None)]):
                headers_cookie_str = str(request.headers.get("Cookie"), encoding="utf-8")
                aiohttp_cookie_dict = ReuseOperation.get_ck_dict_from_headers(headers_ck_str=headers_cookie_str)
                # logger.debug(f"使用 headers 中 ck 的值: {aiohttp_cookie_dict}, {type(aiohttp_cookie_dict)}")
            html_content = await self._request_by_aiohttp(aio_request_args=aiohttp_args_dict, aio_request_cookies=aiohttp_cookie_dict)

        except TimeoutError:
            logger.exception(f'使用 aiohttp 错误响应 url: {request.url}')
            # return self._retry(request, 504, spider)

        # 休眠或延迟设置
        # if not self.sleep:
        if not self.aio_local_conf_exists:
            self.timeout = aiohttp_meta.get("timeout", Param.aiohttp_timeout_default)
            self.sleep = aiohttp_meta.get("sleep", Param.aiohttp_sleep_default)
            self.retry_times = aiohttp_meta.get("retry_times", Param.aiohttp_retry_times_default)
        await asyncio.sleep(self.sleep)

        body = str.encode(html_content)
        if not html_content:
            logger.error(f'url: {request.url} 返回内容为空')

        response = HtmlResponse(
            request.url,
            status=200,
            # headers={"response_header_example": "tmp"},
            body=body,
            encoding='utf-8',
            request=request
        )
        return response

    def process_request(self, request, spider):
        """
        使用 aiohttp 来 process request
        Args:
            request: AiohttpRequest 对象
            spider: scrapy spider

        Returns:
            1). Deferred
        """
        return as_deferred(self._process_request(request, spider))

    async def _spider_closed(self):
        pass

    def spider_closed(self):
        """
        当 spider closed 时调用
        """
        return as_deferred(self._spider_closed())


class AiohttpAsyncMiddleware(AiohttpMiddleware):
    """aiohttp 协程的另一种实现方法，简化 process_request 方法"""

    async def process_request(self, request, spider):
        await super()._process_request(request, spider)



import configparser
from scrapy.commands.genspider import Command
from ayugespidertools.config import NormalConfig


class AyuCommand(Command):

    def short_desc(self):
        return "Print AyugeSpiderTools version"

    def run(self, args, opts):
        config_parse = configparser.ConfigParser()
        config_parse.read(f"{NormalConfig.ROOT_DIR}/pyproject.toml", encoding="utf-8")
        version = config_parse["tool.poetry"]["version"]
        print(f"AyugeSpiderTools {version}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  SqlFormat.py
@Time    :  2022/7/13 11:31
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  sql 相关处理: sql 语句的管理方法，
            这里的 sql 拼接只能做到最简单的逻辑，若想做到适配性更高，请参考 directsql, python-sql, pypika
            或 pymilk 等第三方类似功能库的实现方法，以后会再优化此场景
"""
from typing import Optional, Union


__all__ = [
    "AboutSql",
]


class AboutSql(object):
    """
    这里是生成常用的简单的 sql 语句，如果需要灵活或稍复杂的情况时，请赶快使用 pypika 等成熟的库去！
    """

    @staticmethod
    def select_generate(
        db_table: str,
        key: list,
        rule: dict,
        base: Optional[str] = None,
        order_by: Optional[str] = None,
        limit: Optional[Union[bool, int]] = False
    ) -> (str, tuple):
        """
        根据一些参数来生成供 pymysql 之类的库中使用的 sql 查询语句（适用于简单情况）
        Args:
            db_table: 需要查询的表名称
            key: 需要查询的关键字段
            rule: 查询需要的规则
            base: 在有多个查询规则时，选择 "and" 或 "or"，默认 "and"
            order_by: 排序的 key 值
            limit: limit 限制，默认无限制（查询所有）；如果需要则指定 int 值即可

        Returns:
            select_sql: 生成的 sql 语句
            tuple(rule.values()): 查询字段的参数名称
        """
        select_key = ", ".join('`{}`'.format(k) for k in key)
        select_key = select_key.replace("""`count(*)`""", "count(*)")
        # select_key = select_key.replace("""`count(1)`""", "count(1)")

        base_list = {
            "and": " and ",
            "or": " or ",
        }

        base_where = " and " if not base else base_list[base]
        select_where = base_where.join('`{}`{}%s'.format(k.split('|')[0], k.split('|')[1]) for k in rule.keys())
        if not order_by:
            if limit:
                select_sql = """select %s from `%s` where %s limit %s""" % (select_key, db_table, select_where, limit)
            else:
                select_sql = """select %s from `%s` where %s""" % (select_key, db_table, select_where)
        else:
            select_sql = """select %s from `%s` where %s order by %s desc limit 1""" % (
            select_key, db_table, select_where, order_by)
        """order by id desc"""
        return select_sql, tuple(rule.values())

    @staticmethod
    def insert_generate(db_table: str, data: dict) -> (str, tuple):
        """
        根据一些参数来生成供 pymysql 之类的库中使用的 sql 插入语句
        Args:
            db_table: 需要插入的表名称
            data: 需要插入的关键字段，key: 数据表字段；value: 需插入的参数名

        Returns:
            select_sql: 生成的 sql 语句
            tuple(rule.values()): 新增字段的参数名称
        """
        keys = ", ".join('`{}`'.format(k) for k in data.keys())
        values = ', '.join(['%s'] * len(data))
        sql = '''insert into `%s` (%s) values (%s)''' % (db_table, keys, values)
        return sql, tuple(data.values())

    @staticmethod
    def update_generate(db_table: str, data: dict, rule: dict, base: Optional[str] = None) -> (str, tuple):
        """
        根据一些参数来生成供 pymysql 之类的库中使用的 sql 更新语句
        Args:
            db_table: 需要插入的表名称
            data: 需要更新的 key 和 value 值
            rule: 更新需要的规则
            base: 在有多个查询规则时，选择 "and" 或 "or"，默认 "and"

        Returns:
            select_sql: 生成的 sql 语句
            tuple(rule.values()): 更新字段的参数名称
        """
        update_set = ", ".join('`{}`=%s'.format(k) for k in data.keys())
        base_list = {
            "and": " and ",
            "or": " or ",
        }

        base_where = " and " if not base else base_list[base]
        update_where = base_where.join('`{}`=%s'.format(k) for k in rule.keys())
        sql = """update `%s` set %s where %s""" % (db_table, update_set, update_where)
        return sql, tuple(data.values()) + tuple(rule.values())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  RPA.py
@Time    :  2022/7/15 9:55
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import re
import os
import subprocess


__all__ = [
    "AboutPyppeteer",
]


class AboutPyppeteer(object):
    """
    关于 Pyppeteer 的相关管理，目前只有在 Pyppeteer 卡住的时候及时释放的功能
    TODO: 此功能还不完善，等以后慢慢再优化
    """

    @classmethod
    def get_crawl_debug(cls, content: str) -> int:
        """
        获取日志中的采集信息连续信息的输出次数
        Args:
            content: 在日志文件中读出的内容信息

        Returns:
            len(res): Pyppeteer 连续信息的输出次数
        """
        pattern = re.compile(r"""(\[scrapy\.extensions\.logstats\] INFO: Crawled)""")
        res = pattern.findall(content)
        return len(res)

    @classmethod
    def quit_process(cls, sudo_pwd: str):
        """
        退出程序的方法
        Args:
            sudo_pwd: sudo 需要输入的 root 账号密码

        Returns:
            None
        """
        command = "ps aux | grep run_tmall | grep -v grep | awk '{print $2}' | xargs sudo kill -9"
        os.system('echo %s|sudo -S %s' % (sudo_pwd, command))

        command = "ps aux | grep tmall_by_goods | grep -v grep | awk '{print $2}' | xargs sudo kill -9"
        os.system('echo %s|sudo -S %s' % (sudo_pwd, command))

    @classmethod
    def deal_pyppeteer_suspend(cls, fn: str, line: int):
        """
        将 fn 全路径下的日志中出现 line 次卡住的日志所对应的进程杀掉
        Args:
            fn: 日志的全路径信息
            line: 日志重复输出为 line 次时处理次情况

        Returns:
            None
        """
        read_log = subprocess.getstatusoutput('tail -n %d %s' % (line, fn))

        # 当运行成功时
        if not read_log[0]:
            log_res = ''.join(read_log[1])

            block_times = cls.get_crawl_debug(content=log_res)
            # 当连续输出 scrapy 的统计信息 3 次时，则卡住
            if block_times >= 3:
                # logger.info("quit process success")
                cls.quit_process("自行替换 sudo root 的密码")

            # 当最新四行日志中未出现 scrapy 统计，则为正常状态，并清空日志
            elif block_times == 0:
                clean_log = subprocess.getstatusoutput('> %s' % (fn,))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
import pymysql
from typing import Optional


__all__ = [
    'MysqlOrm',
]


class MysqlOrm(object):
    """数据库的简单使用，结合 SqlFormat 方法使用（临时使用）"""

    def __init__(self, pymsql_connect_config: dict):
        self.connection = pymysql.connect(**pymsql_connect_config)
        self.cursor = self.connection.cursor()

    def insert_data(self, sql_pre: str, sql_after: tuple):
        self.connection.ping(reconnect=True)
        self.cursor.execute(sql_pre, sql_after)
        self.connection.commit()

    def search_data(self, sql_pre: str, sql_after: tuple, type: Optional[str] = None) -> (bool, tuple):
        if not type:
            type = "one"

        self.connection.ping(reconnect=True)
        self.cursor.execute(sql_pre, sql_after)
        global select_res
        if type == "one":
            select_res = self.cursor.fetchone()

        elif type == "all":
            select_res = self.cursor.fetchall()

        # 判断查询结果
        if select_res:
            return True, select_res
        return False, ""

    def update_data(self, sql_pre: str, sql_after: tuple):
        self.connection.ping(reconnect=True)
        self.cursor.execute(sql_pre, sql_after)
        self.connection.commit()

    def close(self):
        self.cursor.close()
        self.connection.close()

    def __del__(self):
        self.cursor.close()
        self.connection.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ayugespidertools',
 'ayugespidertools.commands',
 'ayugespidertools.common',
 'ayugespidertools.templates.project.module',
 'ayugespidertools.templates.project.module.common',
 'ayugespidertools.templates.project.module.spiders',
 'ayugespidertools.utils']

package_data = \
{'': ['*'],
 'ayugespidertools': ['VIT/*',
                      'docs/*',
                      'templates/project/*',
                      'templates/spiders/*'],
 'ayugespidertools.templates.project.module': ['VIT/*']}

install_requires = \
['DBUtils>=3.0.2,<4.0.0',
 'Pillow>=9.2.0,<10.0.0',
 'PyExecJS>=1.5.1,<2.0.0',
 'PyMySQL>=1.0.2,<2.0.0',
 'SQLAlchemy>=1.4.41,<2.0.0',
 'Scrapy>=2.6.3,<3.0.0',
 'WorkWeixinRobot>=1.0.1,<2.0.0',
 'aiohttp>=3.8.3,<4.0.0',
 'environs>=9.5.0,<10.0.0',
 'html2text>=2020.1.16,<2021.0.0',
 'itemadapter>=0.7.0,<0.8.0',
 'loguru>=0.6.0,<0.7.0',
 'numpy>=1.23.3,<2.0.0',
 'opencv-python>=4.6.0.66,<5.0.0.0',
 'pandas>=1.5.0,<2.0.0',
 'pymongo>=3.12.3,<4.0.0',
 'requests>=2.28.1,<3.0.0',
 'retrying>=1.3.3,<2.0.0']

entry_points = \
{'console_scripts': ['ayugespidertools = '
                     'ayugespidertools.utils.cmdline:execute']}

setup_kwargs = {
    'name': 'ayugespidertools',
    'version': '1.0.5',
    'description': 'scarpy 扩展库：在爬虫和其它开发中遇到的一些工具库集合',
    'long_description': '# AyugeSpiderTools 工具说明\n\n> 本文章用于说明在爬虫开发中遇到的各种通用方法，还有 `scrapy` 扩展库开发，将其打包成 `Pypi` 包以方便安装和使用，此工具会长久维护。\n\n## 前言\n在 `Python` 开发时会遇到一些经常使用的自封装模块，为了更方便地使用这些模块，我们需要将其打包并发布到 `Pypi` 上。\n\n## 项目状态\n\n> 目前项目正处于**积极开发和维护**中\n\n项目创建初期，且包含各阶段时期的代码，会出现代码风格不统一，代码杂乱，文档滞后和测试文件不全的问题，近期会优化此情况。\n\n项目目前暂定主要包含**两大部分**（*可能其中一些功能会考虑拆分到其它新建库中，以防止此库过于杂糅*）：\n\n- 开发场景中的工具库\n  - 比如 `MongoDB`，`Mysql sql` 语句的生成，图像处理，数据处理相关 ... ...\n- `Scrapy` 扩展功能\n  - 使爬虫开发无须在意数据库和数据表结构，不用去管常规 `item, pipelines` 和 `middlewares` 的文件的编写\n\n注：具体内容请查看本文中的 [TodoList](# TodoList) 内容\n\n## 1. 前提条件\n\n> `python 3.8+` 可以直接输入以下命令：\n\n```shell\npip install ayugespidertools -i https://pypi.org/simple\n```\n\n> 或者自行安装相关依赖：\n\n依赖信息在根目录下的 `requirements.txt` 文件中\n\n注：本库依赖中的 `pymongo` 版本要在 `3.12.3` 及以下是因为我的 `mongoDB` 的版本为 `3.4`，`pymogo` 官方从 `3.12.3` 以后的版本开始不再支持 `3.6` 版本以下的 `MongoDB` 数据库了，望周知！；\n建议在独立的虚拟环境中使用，以免与您的其它项目依赖互相干扰。\n\n### 1.1. 使用方法\n\n#### 1.1.1. Scrapy 扩展库场景\n\n> 此扩展使 `Scrapy` 爬虫开发不用考虑其 `item` 编写，内置通用的 `middlewares` 中间件方法（随机请求头，动态/独享代理等），和常用的 `pipelines` 方法（`Mysql`，`MongoDB` 存储，`Kafka`，`RabbitMQ` 推送队列等）。\n>\n> 开发人员只需根据命令生成示例模板，再配置并激活相关设置即可，可以专注于爬虫 `spider` 的开发。\n\n具体使用方法请在 [DemoSpider 之 AyugeSpiderTools 工具应用示例](https://github.com/shengchenyang/DemoSpider) 项目中查看，目前已适配以下场景：\n\n```diff\n# 采集数据存入 `Mysql` 的场景：\n- 1).demo_one: 配置根据本地 `settings` 的 `LOCAL_MYSQL_CONFIG` 中取值\n+ 3).demo_three: 配置根据 `consul` 的应用管理中心中取值\n+ 5).demo_five: 异步存入 `Mysql` 的场景\n\n# 采集数据存入 `MongoDB` 的场景：\n- 2).demo_two: 采集数据存入 `MongoDB` 的场景（配置根据本地 `settings` 的 `LOCAL_MONGODB_CONFIG` 中取值）\n+ 4).demo_four: 采集数据存入 `MongoDB` 的场景（配置根据 `consul` 的应用管理中心中取值）\n+ 6).demo_six: 异步存入 `MongoDB` 的场景\n\n# 将 `Scrapy` 的 `Request`，`FormRequest` 替换为其它工具实现的场景\n- 以上为使用 scrapy Request 的场景\n+ 7).demo_seven: scrapy Request 替换为 requests 请求的场景(一般情况下不推荐使用，同步库会拖慢 scrapy 速度，可用于测试场景)\n\n+ 8).demo_eight: 同时存入 Mysql 和 MongoDB 的场景\n\n- 9).demo_aiohttp_example: scrapy Request 替换为 aiohttp 请求的场景，提供了各种请求场景示例（GET,POST）\n+ 10).demo_aiohttp_test: scrapy aiohttp 在具体项目中的使用方法示例\n\n+ 11).demo_proxy_one: 快代理动态隧道代理示例\n+ 12).demo_proxy_two: 测试快代理独享代理\n```\n\n注：具体内容及时效性请以 [DemoSpider](https://github.com/shengchenyang/DemoSpider) 项目中描述为准。\n\n#### 1.1.2. 开发场景\n\n其开发场景下的功能，请在本文 [2. 功能介绍](# 2. 功能介绍) 部分中查看。\n\n## 2. 功能介绍\n\n### 2.1. 数据格式化\n\n> 目前此场景下的功能较少，后面会慢慢丰富其功能\n\n#### 2.1.1. get_full_url\n\n根据域名 `domain_name` 拼接 `deal_url` 来获得完整链接，示例如下：\n\n```python\nfull_url = FormatData.get_full_url(domain_name="https://static.geetest.com", deal_url="/captcha_v3/batch/v3/2021-04-27T15/word/4406ba6e71cd478aa31e0dca37601cd4.jpg")\n```\n\n输出为：\n\n```\nhttps://static.geetest.com/captcha_v3/batch/v3/2021-04-27T15/word/4406ba6e71cd478aa31e0dca37601cd4.jpg\n```\n\n#### 2.1.2. click_point_deal\n\n将小数 `decimal` 保留小数点后 `decimal_places` 位，结果四舍五入，示例如下：\n\n```\nres = FormatData.click_point_deal(13.32596516, 3)\n```\n\n输出为：\n\n```\n13.326\n```\n\n#### 2.1.3. normal_to_stamp\n\n将网页中显示的正常时间转为时间戳\n\n```python\nnormal_stamp = FormatData.normal_to_stamp("Fri, 22 Jul 2022 01:43:06 +0800")\nprint("normal_stamp1:", normal_stamp)\n\nnormal_stamp = FormatData.normal_to_stamp("Thu Jul 22 17:59:44 2022")\nprint("normal_stamp2:", normal_stamp)\n\nnormal_stamp = FormatData.normal_to_stamp("2022-06-21 16:40:00")\n\nnormal_stamp = FormatData.normal_to_stamp("2022/06/21 16:40:00")\nprint("normal_stamp4:", normal_stamp)\n\nnormal_stamp = FormatData.normal_to_stamp("2022/06/21", date_is_full=False)\nprint("normal_stamp4_2:", normal_stamp)\n\n# 当是英文的其他格式，或者混合格式时，需要自己自定时间格式化符\nnormal_stamp = FormatData.normal_to_stamp(normal_time="2022/Dec/21 16:40:00", _format_t="%Y/%b/%d %H:%M:%S")\nprint("normal_stamp5:", normal_stamp)\n```\n\n输出为：\n\n```\nnormal_stamp1: 1658425386\nnormal_stamp2: 1658483984\nnormal_stamp3: 1655800800\nnormal_stamp4: 1655800800\nnormal_stamp4_2: 1655740800\nnormal_stamp5: 1671612000\n```\n\n### 2.2. 图片相关操作\n\n#### 2.2.1. 滑块验证码缺口距离识别\n\n通过背景图片和缺口图片识别出滑块距离，示例如下：\n\n```python\n# 参数为图片全路径的情况\ngap_distance = Picture.identify_gap("doc/image/2.jpg", "doc/image/1.png")\nprint("滑块验证码的缺口距离1为：", gap_distance)\nassert gap_distance in list(range(205, 218))\n\n# 参数为图片 bytes 的情况\nwith open("doc/image/1.png", "rb") as f:\n    target_bytes = f.read()\nwith open("doc/image/2.jpg", "rb") as f:\n    template_bytes = f.read()\ngap_distance = Picture.identify_gap(template_bytes, target_bytes, "doc/image/33.png")\nprint("滑块验证码的缺口距离2为：", gap_distance)\n```\n\n结果为：\n\n<img src="http://175.178.210.193:9000/drawingbed/image/image-20220802110652131.png" alt="image-20220802110652131" style="zoom: 25%;" />\n\n<img src="http://175.178.210.193:9000/drawingbed/image/image-20220802110842737.png" alt="image-20220802110842737" style="zoom:33%;" />\n\n#### 2.2.2. 滑块验证轨迹生成\n\n根据滑块缺口的距离生成轨迹数组，目前也不是通用版。\n\n```python\ntracks = VerificationCode.get_normal_track(space=120)\n```\n\n结果为：\n\n```\n生成的轨迹为： [[2, 2, 401], [4, 4, 501], [8, 6, 603], [13, 7, 701], [19, 7, 801], [25, 7, 901], [32, 10, 1001], [40, 12, 1101], [48, 14, 1201], [56, 15, 1301], [65, 18, 1401], [74, 19, 1501], [82, 21, 1601], [90, 21, 1701], [98, 22, 1801], [105, 23, 1901], [111, 25, 2001], [117, 26, 2101], [122, 28, 2201], [126, 30, 2301], [128, 27, 2401], [130, 27, 2502], [131, 30, 2601], [131, 28, 2701], [120, 30, 2802]]\n```\n\n### 2.3. Mysql 相关\n\n`sql` 语句简单场景生成，目前是残废版，只适用于简单场景。\n\n更多复杂的场景请查看 [directsql](https://pypi.org/project/directsql/#history), [python-sql](https://pypi.org/project/python-sql/#history),  [pypika](https://pypi.org/project/PyPika/#description) 或 [pymilk](https://pypi.org/project/pymilk/) 的第三方库实现，以后会升级本库的方法。\n\n```python\n# mysql 连接\nfrom ayugespidertools import MysqlClient\nfrom ayugespidertools.common.SqlFormat import AboutSql\n\n\nmysql_client = MysqlClient.MysqlOrm(NormalConfig.PYMYSQL_CONFIG)\n\n# test_select_data\nselect_sql, select_value = AboutSql.select_generate(db_table="zhihu_answer_info", key=["id", "q_title"], rule={"q_id|=": "34987206"}, limit=1)\nprint(f"select_sql: {select_sql}, select_value: {select_value}")\nmysql_client.search_data(select_sql, select_value, type="one")\n\n# test_insert_data\ninsert_sql, insert_value = AboutSql.insert_generate(db_table="user", data={"name": "zhangsan", "age": 18})\nprint(f"insert_sql: {insert_sql}, insert_value: {insert_value}")\nmysql_client.insert_data(insert_sql, insert_value)\n\n# test_update_data\nupdate_sql, update_value = AboutSql.update_generate(db_table="user", data={"score": 4}, rule={"name": "zhangsan"})\nprint(f"update_sql: {update_sql}, update_value: {update_value}")\nmysql_client.update_data(update_sql, update_value)\n```\n\n结果为：\n```\nselect_sql: select `id`, `q_title` from `zhihu_answer_info` where `q_id`=%s limit 1, select_value: (\'34987206\',)\n\ninsert_sql: insert into `user` (`name`, `age`) values (%s, %s), insert_value: (\'zhangsan\', 18)\n\nupdate_sql: update `user` set `score`=%s where `name`=%s, update_value: (4, \'zhangsan\')\n```\n\n### 2.4. 自动化相关\n\n目前是残废阶段，以后放上一些自动化相关操作\n\n### 2.5. 执行 js 相关\n\n鸡肋封装，以后会优化和添加多个常用功能\n\n```python\n# 测试运行 js 文件中的方法\njs_res = RunJs.exec_js("doc/js/add.js", "add", 1, 2)\nprint("test_exec_js:", js_res)\nassert js_res\n\n# 测试运行 ctx 句柄中的方法\nwith open(\'doc/js/add.js\', \'r\', encoding=\'utf-8\') as f:\n    js_content = f.read()\nctx = execjs.compile(js_content)\n\njs_res = RunJs.exec_js(ctx, "add", 1, 2)\nprint("test_exec_js_by_file:", js_res)\nassert js_res\n```\n\n## 3. 总结\n\n项目目前是疯狂开发阶段，会慢慢丰富 `python` 开发中的遇到的通用方法。\n\n## TodoList\n\n- [x] `scrapy` 的扩展功能场景\n  - [ ] `scrapy` 结合 `crawlab` 的日志统计功能\n  - [x] `scrapy` 脚本运行信息统计和项目依赖表采集量统计，可用于日志记录和预警\n  - [x] 自定义模板，在 `ayugespidertools startproject <projname>` 和 `ayugespidertools genspider <spidername>` 时生成适合本库的模板文件\n  - [x] ~~增加根据 `nacos` 来获取配置的功能~~ -> 改为增加根据 `consul` 来获取配置的功能\n  - [x] 代理中间件（独享代理、动态隧道代理）\n  - [x] 随机请求头 `UA` 中间件，根据 `fake_useragent` 中的权重来随机\n  - [x] 使用以下工具来替换 `scrapy` 的 `Request` 来发送请求\n    - [ ] `selenum`: 性能没有 `pyppeteer` 强\n    - [x] `pyppeteer`: `Gerapy-pyppeteer` 库已经实现此功能\n    - [x] `requests`: 这个不推荐使用，`requests` 同步库会降低 `scrapy` 运行效率\n    - [ ] `splash`: 继承 `splash` 渲染 `js` 的功能\n    - [x] `aiohttp`: 集成将 `scrapy Request` 替换为 `aiohttp` 的协程方式\n  - [x] `Mysql` 存储的场景下适配\n    - [x] 自动创建 `Mysql` 用户场景下需要的数据库和数据表及字段格式，还有字段注释\n  - [x] `MongoDB` 存储的场景下适配，编写风格与 `Mysql` 存储等场景下一致\n  - [ ] 集成 `Kafka`，`RabbitMQ` 等数据推送功能\n  - [ ] ... ...\n- [x] 常用开发场景\n  - [x] `sql` 语句拼接，只是简单场景，后续优化。已给出优化方向，参考库等信息。\n  - [x] `mongoDB` 语句拼接\n  - [x] 数据格式化处理，比如：去除网页标签，去除无效空格等\n  - [ ] 字体反爬还原方法\n  - [ ] `html` 格式转 `markdown` 格式\n  - [ ] `html` 数据处理，去除标签，不可见字符，特殊字符改成正常显示等等等\n  - [x] 添加常用的图片验证码中的处理方法\n    - [x] 滑块缺口距离的识别方法（多种实现方式）\n    - [x] 根据滑块距离生成轨迹数组的方法\n    - [x] 识别点选验证码位置及点击顺序，识别结果不太好，待优化\n    - [ ] 图片乱序混淆的还原方法\n  - [ ] ... ...\n',
    'author': 'ayuge',
    'author_email': 'ayuge.s@qq.com',
    'maintainer': 'ayuge',
    'maintainer_email': 'ayuge.s@qq.com',
    'url': 'http://175.178.210.193:8090/mkdocs-material/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  RunJs.py
@Time    :  2022/7/8 14:55
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import execjs


__all__ = [
    'exec_js',
]


def exec_js(js_handle, method, *args):
    """
    执行 js_full_path 的 js 文件中的 method 方法
    Args:
        js_handle: js 处理对象，可能是 js_full_path 的全路径，也可能是 js 的 ctx 句柄
        method: 需要调用的 js 方法名
        *args: 调用的目标函数所需要的参数

    Returns:
        js_res: js 执行后的结果
    """
    # 先判断是否有 ctx 对象，如果有的话，直接调用 ctx 的方法即可，不用重复构造 ctx 对象
    if any([type(js_handle) != str, type(js_handle) == execjs._external_runtime.ExternalRuntime.Context]):
        js_res = js_handle.call(method, *args)

    # 没有 ctx 句柄则需要创建此对象
    else:
        with open(js_handle, 'r', encoding='utf-8') as f:
            js_content = f.read()
        ctx = execjs.compile(js_content)
        js_res = ctx.call(method, *args)
    return js_res
